"""Domain-Driven Design Aggregate Root."""
import abc

import pydantic

from .entity import Entity
from .event import Event


class AggregateRoot(Entity, abc.ABC):
    """Special type of Entity that is the root of an Aggregate of domain objects.

    Provides a clear and controlled entry point for managing and interacting with related domain objects,
    enforcing invariants, and ensuring data consistency.

    An Aggregate is a cluster of domain objects that are treated as a single unit for data changes.
    It's a way of grouping related entities and value objects together.
    Aggregates ensure that the encapsulated objects are only modified through well-defined and controlled access points.
    Aggregates help maintain the integrity of the domain model and represent consistency boundaries.

    Within an Aggregate, one specific entity is designated as the "Aggregate Root."
    The Aggregate Root is the entry point for any operations or modifications within the Aggregate.
    All external interactions, such as client code or application services,
    should go through the Aggregate Root to access or change the data within the Aggregate.
    The Aggregate Root is responsible for maintaining the integrity of the entire Aggregate by enforcing business rules and invariants.

    Attributes:
        _events: A list of domain events associated with the aggregate.
    """

    _events: list[Event] = pydantic.PrivateAttr(default_factory=list)

    def add_event(self, event: Event) -> None:
        self._events.append(event)
        return None

    def add_events(self, events: list[Event]) -> None:
        self._events.extend(events)
        return None

    def get_events(self) -> list[Event]:
        return self._events

    def clear_events(self) -> None:
        self._events.clear()
        return None



import datetime

import pydantic
import pytest

import pydoca


class FakeEvent(pydoca.Event):
    attribute1: int
    attribute2: str


def test_event_immutable() -> None:
    event = FakeEvent(attribute1=10, attribute2="test")
    with pytest.raises(pydantic.ValidationError):
        event.attribute1 = 20


def test_event_equality() -> None:
    now = datetime.datetime.now(datetime.timezone.utc)
    event1 = FakeEvent(attribute1=10, attribute2="test", timestamp=now)
    event2 = FakeEvent(attribute1=10, attribute2="test", timestamp=now)
    event3 = FakeEvent(attribute1=20, attribute2="test")

    assert event1 == event2  # Same attributes, so they should be equal
    assert event1 != event3  # Different attribute value


def test_event_hash() -> None:
    now = datetime.datetime.now(datetime.timezone.utc)
    event1 = FakeEvent(attribute1=10, attribute2="test", timestamp=now)
    event2 = FakeEvent(attribute1=10, attribute2="test", timestamp=now)
    event3 = FakeEvent(attribute1=10, attribute2="test")  # Different timestamp

    assert hash(event1) == hash(
        event2
    )  # Same attributes, so their hash values should be equal
    assert hash(event1) != hash(event3)  # Different attribute value


def test_event_extra_attributes() -> None:
    with pytest.raises(pydantic.ValidationError):
        FakeEvent(attribute1=10, attribute2="test", extra_attribute=30)  # type: ignore[call-arg]


def test_direct_use_of_event() -> None:
    with pytest.raises(TypeError):
        pydoca.Event()


def test_event_timestamp_default() -> None:
    event = FakeEvent(attribute1=10, attribute2="test")
    now = datetime.datetime.now(datetime.timezone.utc)
    assert (
        event.timestamp <= now
    )  # The timestamp should be less than or equal to the current time



import abc
import functools
import inspect
import itertools
from types import MappingProxyType
from typing import Any, Callable, Generic, Optional, Self, TypeVar, cast

from .aggregate_root import AggregateRoot
from .event import Event
from .port_adapter import Port

AggregateRootT = TypeVar("AggregateRootT", bound=AggregateRoot)


class Session(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def start(cls) -> Self:
        """Starts the session."""

    @classmethod
    @abc.abstractmethod
    def url(cls) -> str:
        """Returns the session url."""

    @abc.abstractmethod
    def commit(self) -> None:
        """Commits the session."""

    @abc.abstractmethod
    def rollback(self) -> None:
        """Rollbacks changes if any."""

    def __eq__(self, other: Any) -> bool:
        """Compares two sessions."""
        return isinstance(other, self.__class__) and self.url() == other.url()


TWrap = TypeVar("TWrap", bound=Callable[..., Any])


class Repository(Generic[AggregateRootT], Port):
    """Repository interface."""

    sessionT: type[Session]
    events: list[Event]
    _session: Optional[Session] = None

    def __init__(self) -> None:
        self.events = []

    @property
    def session(self) -> Session:
        if self._session:
            return self._session
        self._session = self.sessionT.start()
        return self._session

    def set_session(self, session: Session) -> None:
        self._session = session

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)  # Call for Port
        if isinstance(cls, abc.ABC):
            return  # Assumes every non ABC class is a real repository implementation
        for name, fn in inspect.getmembers(cls, inspect.isfunction):
            if name.startswith("__") and name.endswith("__"):
                continue
            parameters: MappingProxyType[str, inspect.Parameter] = inspect.signature(
                fn
            ).parameters
            for parameter in parameters.values():
                if inspect.isclass(parameter.annotation) and issubclass(
                    parameter.annotation, AggregateRoot
                ):
                    setattr(cls, name, cls.track_events(fn))

    @classmethod
    def track_events(cls, func: TWrap) -> TWrap:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            for param in itertools.chain(args, kwargs.values()):
                if isinstance(param, AggregateRoot):
                    args[0].events.extend(param.get_events())
                    param.clear_events()
                if isinstance(param, list):
                    for elem in param:
                        if isinstance(elem, AggregateRoot):
                            args[0].events.extends(elem.get_events())
                            elem.clear_events()
            return func(*args, **kwargs)

        return cast(TWrap, wrapper)



import decimal

import pydantic

import pydoca

from .. import domain
from . import _ports


class Operation(pydoca.ValueObject):
    source: str
    frequency: domain.FrequencyPerYear
    amount: decimal.Decimal
    type: str  # income or expense

    @pydantic.field_validator("type")
    @classmethod
    def check_type_value(cls, v: str) -> str:
        if v not in {"income", "expense"}:
            raise ValueError("Budget operation must be income or expense")
        return v


class AddToBudgetCmd(pydoca.Command):
    budget_id: str
    operations: list[Operation] = []


class AddToBudget(pydoca.UseCase):
    class UnitOfWork:
        budget_repo: _ports.BudgetRepository

    def exec(self, cmd: AddToBudgetCmd) -> domain.Budget:
        with self.uow as uow:
            budget: domain.Budget = uow.budget_repo.get_by_id(cmd.budget_id)
            for ope in cmd.operations:
                if ope.type == "income":
                    budget.add_income(
                        source=ope.source, frequency=ope.frequency, amount=ope.amount
                    )
                if ope.type == "expense":
                    budget.add_expense(
                        source=ope.source, frequency=ope.frequency, price=ope.amount
                    )
            return budget



import collections.abc
from typing import Iterator, Self

import pydoca
import tests.integration.test_app.budget.application as application
import tests.integration.test_app.budget.domain as domain

db = {}


class InMemorySession(pydoca.Session, collections.abc.MutableMapping):
    def __init__(self):
        self.store: dict[str, domain.Budget] = db

    def __setitem__(self, key: str, val: domain.Budget) -> None:
        self.store[key] = val

    def __delitem__(self, key: str) -> None:
        del self.store[key]

    def __getitem__(self, key: str) -> domain.Budget:
        return self.store[key]

    def __len__(self) -> int:
        return len(self.store)

    def __iter__(self) -> Iterator[str]:
        return iter(self.store)

    @classmethod
    def start(cls) -> Self:
        return cls()

    @classmethod
    def url(cls) -> str:
        return "//memory"

    def commit(self) -> None:
        print("Commit")

    def rollback(self) -> None:
        print("Rollback")


class InMemoryBudgetRepo(application.BudgetRepository):
    sessionT = InMemorySession

    def get_by_id(self, budget_id: str) -> domain.Budget:
        if budget := self.session.get(budget_id):
            return budget
        else:
            raise pydoca.EntityNotFoundError(class_id=(domain.Budget, budget_id))

    def save(self, budget: domain.Budget, create: bool = False) -> domain.Budget:
        if create and budget.id in self.session:
            raise pydoca.EntityAlreadyExistError(entity=budget)
        self.session[budget.id] = budget
        return budget



"""Very simplistic domain representing a budget based on incomes and expenses."""
import decimal
import enum

import pydoca


class Currency(str, enum.Enum):
    """ISO 4217 currency codes."""

    CAD = "Canadian dollar"
    EUR = "Euro"
    USD = "United States dollar"


class FrequencyPerYear(enum.Enum):
    YEARLY = 1
    MONTHLY = 12
    SEMI_MONTHLY = 24
    BI_WEEKLY = 26
    WEEKLY = 52
    DAILY = 365


class Income(pydoca.Entity):
    """Money a person or entity receives in exchange for their labor or investment."""

    source: str
    frequency: FrequencyPerYear
    amount: decimal.Decimal

    def _id(self) -> str:
        return self.source.lower()


class Expense(pydoca.Entity):
    """Payment for an item, service, or other category of costs."""

    source: str
    frequency: FrequencyPerYear
    price: decimal.Decimal

    def _id(self) -> str:
        return self.source.lower()


class IncomeAdded(pydoca.Event):
    source: str


class ExpenseAdded(pydoca.Event):
    source: str


class Budget(pydoca.AggregateRoot):
    title: str
    currency: Currency
    incomes: list[Income] = []
    expenses: list[Expense] = []

    def _id(self) -> str:
        return self.title.lower()

    def add_income(
        self, source: str, frequency: FrequencyPerYear, amount: decimal.Decimal
    ) -> None:
        income = Income(source=source, frequency=frequency, amount=amount)
        if income in self.incomes:
            raise pydoca.EntityAlreadyExistError(income)
        self.incomes.append(income)
        self.add_event(IncomeAdded(source=source))

    def add_expense(
        self, source: str, frequency: FrequencyPerYear, price: decimal.Decimal
    ) -> None:
        expense = Expense(source=source, frequency=frequency, price=price)
        if expense in self.expenses:
            raise pydoca.EntityAlreadyExistError(expense)
        self.expenses.append(expense)
        self.add_event(ExpenseAdded(source=source))

    def calculate_cash_flow_per_month(self) -> decimal.Decimal:
        total = decimal.Decimal(0)

        for income in self.incomes:
            total += income.amount * income.frequency.value / 12

        for expense in self.expenses:
            total -= expense.price * expense.frequency.value / 12

        return total



from typing import Optional

from .port_adapter import AdaptersConfig


def bootstrap(adapters_config: Optional[type[AdaptersConfig]] = None) -> None:
    if adapters_config:
        adapters_config()
    return None



import abc

import pytest

import pydoca


class EmailService(pydoca.Service):
    @abc.abstractmethod
    def send_email(self, message: str) -> None:
        """Sends an email."""


class FakeEmailService(EmailService):
    def send_email(self, message: str) -> None:
        print(message)
        return


def test_port_registry():
    # Declaring EmailService class should register it
    assert pydoca.Port._registry["EmailService"] == EmailService


def test_factory_bind():
    pydoca.bind(EmailService, FakeEmailService)
    assert isinstance(pydoca.inject(EmailService), FakeEmailService)


def test_adapter_bind():
    # Test provider bind and configuration override
    email_svc = FakeEmailService()
    pydoca.bind(EmailService, email_svc)
    assert pydoca.inject(EmailService) == email_svc


def test_clear_and_error():
    pydoca.bind(EmailService, FakeEmailService)
    pydoca.clear()

    with pytest.raises(
        pydoca.AdapterNotConfiguredError,
        match="Adapter for EmailService port not configured",
    ):
        pydoca.inject(EmailService)



"""Port/Adapter architecture pattern."""
import abc
import inspect
import logging
from typing import Any, Callable, ClassVar, Optional, Self

logger = logging.getLogger(__name__)

PortClassName = str


class Port(abc.ABC):
    """Port.

    Abstract classes inheriting Port are considered Port in the Port/Adapter pattern.
    They should be defined in the application layer of your project. One layer above your domain
    and used in your Use Cases.

    Classes inheriting and implementing those abstract classes are considered the adapters,
    and should be defined in the adapters, outer layer of your project, alongside the actors.

    Attributes:
        _registry: Keeps track of the Port classes so we can find them using only their names.
    """

    _registry: ClassVar[dict[PortClassName, type[Self]]] = {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if cls.__name__ in ["Repository", "Service"] or not inspect.isabstract(cls):
            # Do not register our bases Repository and Service abc classes, they are part of pydoca.
            # Do not register real implementations (adapters) to the ports registry as there
            # __init_subclass__ is also triggered here.
            return

        super().__init_subclass__(**kwargs)
        # Should be an abstract class inheriting `Port` declared as part of the user application ports.
        cls._registry[cls.__name__] = cls
        logger.info(f"Register {cls.__name__} port")


PortType = type[Port]


class AdapterNotConfiguredError(Exception):
    def __init__(self, port: PortType) -> None:
        super().__init__(f"Adapter for {port.__name__} port not configured")


class PortNotFoundError(Exception):
    def __init__(self, port: PortClassName) -> None:
        super().__init__(f"Port class {port} not found.")


Adapter = object | Any
AdapterFactory = Callable[[], Adapter]

_ADAPTERS_CONFIGURATION: dict[PortType, AdapterFactory | Adapter] = {}


def bind(port: PortType, adapter: AdapterFactory | Adapter) -> None:
    _ADAPTERS_CONFIGURATION[port] = adapter
    logger.info(f"Bind {port} port to {adapter} adapter")


def clear() -> None:
    _ADAPTERS_CONFIGURATION.clear()


def inject(port: PortType) -> Adapter:
    adapter: Optional[AdapterFactory | Adapter] = _ADAPTERS_CONFIGURATION.get(port)
    if not adapter:
        raise AdapterNotConfiguredError(port)

    if callable(adapter):
        return adapter()
    else:
        return adapter


class AdaptersConfig:
    """Adapters configuration to automatically bind them when bootstrap.

    class DevConfiguration(pydoca.AdaptersConfig):
        YourRepository = adapters.SQLiteRepo
        YourService = adapters.FakeService

    pydoca.bootstrap(adapters_config=DevConfiguration)

    This configuration alongside bootstrap will bind YourRepository port to SQLiteRepo adapter and
    YourService port to FakeService adapter.
    """

    def __init__(self) -> None:
        for port_name, adapter_class in [
            (key, val)
            for key, val in self.__class__.__dict__.items()
            if not key.startswith("__")
        ]:
            if port_name in Port._registry:
                bind(Port._registry[port_name], adapter_class)
            else:
                raise PortNotFoundError(port_name)

    def __init_subclass__(cls, **kwargs: Any):
        # Forces subclass to trigger super init
        cls.__init__ = lambda self: super(cls, self).__init__()  # type: ignore[method-assign]



import decimal

import tests.integration.test_app.budget.application as application
import tests.integration.test_app.budget.domain as domain


class FakeExchangeRateSvc(application.ExchangeRateService):
    def convert_currency(
        self,
        amount: decimal.Decimal,
        current_currency: domain.Currency,
        new_currency: domain.Currency,
    ) -> decimal.Decimal:
        if current_currency is not domain.Currency.CAD:
            raise ValueError(f"Can not convert from {current_currency.value}")

        rates_to_usd = {
            "CAD": 1,
            "EUR": 0.68,
            "USD": 0.75,
        }

        return amount * decimal.Decimal(rates_to_usd[new_currency.name])



import logging
import queue
from contextvars import ContextVar
from types import TracebackType
from typing import Any, Iterator, Optional, Self

import pydantic

from .event import Event
from .port_adapter import inject
from .repository import Repository, Session

logger = logging.getLogger(__name__)

_EVENT_BUS: ContextVar[queue.Queue[Event]] = ContextVar(
    "EVENT_BUS", default=queue.Queue()
)


class EventBus:
    @staticmethod
    def publish_events(events: Iterator[Event]) -> None:
        try:
            event_bus = _EVENT_BUS.get()
            for event in events:
                event_bus.put(event)
        except Exception:
            logger.exception(f"Error publishing events {events}")

    @staticmethod
    def get_event() -> Optional[Event]:
        try:
            return _EVENT_BUS.get().get_nowait()
        except queue.Empty:
            return None

    @staticmethod
    def get_event_block() -> Event:
        return _EVENT_BUS.get().get()


class NotARepositoryError(Exception):
    """If the object is not a repository."""


class DifferentSessionsError(Exception):
    """If the unit of work detects different sessions."""


class UnitOfWorkBase(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    # TODO: Manage multiple sessions
    _session: Optional[Session] = None

    @property
    def session(self) -> Session:
        if not self._session:
            raise RuntimeError("UnitOfWork session not set.")
        return self._session

    @property
    def repositories(self) -> Iterator[Repository[Any]]:
        for val in dict(self).values():
            if isinstance(val, Repository):
                yield val

    @pydantic.model_validator(mode="before")
    @classmethod
    def inject_repositories(cls, data: dict[str, Repository[Any]]) -> Any:
        if not data:
            data = {}
        for attr_name, attr_info in cls.model_fields.items():
            repository = inject(attr_info.annotation)  # type: ignore
            if not isinstance(repository, Repository):
                raise NotARepositoryError(
                    f"{repository.__class__.__name__} is not a Repository."
                )
            data[attr_name] = repository
        return data

    def __enter__(self) -> Self:
        repo: Repository[Any]
        for repo in dict(self).values():
            session = repo.session
            if not self._session:
                self._session = session
                continue
            if session != self._session:
                raise DifferentSessionsError(
                    "UnitOfWork can not manage different sessions."
                )
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        if exc_type:
            self.session.rollback()
        else:
            self.commit()

    def commit(self) -> None:
        try:
            self.session.commit()
        except Exception as exc:
            # Collect the events the clear the Aggregate but do not publish them
            self.collect_events()
            logger.exception(f"Error while committing the session: {exc}")
            raise exc
        else:
            EventBus.publish_events(self.collect_events())

    def collect_events(self) -> Iterator[Event]:
        for repo in self.repositories:
            for i in range(len(repo.events)):
                yield repo.events[i]



"""Domain-Driven Design Value Object."""
from typing import Any

import pydantic


class ValueObject(pydantic.BaseModel):
    """Immutable objects with no individual identity.

    Value objects are immutable and represent a concept in the domain
    that is defined by its attributes rather than its identity. They
    have no unique identifier and are typically used to encapsulate a
    group of related attributes or properties.
    They compare equality against their attributes rather than their identities.
    """

    model_config = pydantic.ConfigDict(frozen=True, extra="forbid")

    def __init__(self, **data: Any) -> None:
        if type(self) is ValueObject:
            raise TypeError(
                "ValueObject cannot be instantiated directly. Please subclass it and define your attributes."
            )
        super().__init__(**data)



import decimal

import pydoca

from .. import domain
from . import _ports


class CalculateCashflowCmd(pydoca.Command):
    budget_id: str
    currency: domain.Currency


class CalculateCashFlow(pydoca.UseCase):
    # Test simple service injection
    exchange_rate_svc: _ports.ExchangeRateService

    # Test repo injection without uow
    budget_repo: _ports.BudgetRepository

    def exec(self, cmd: CalculateCashflowCmd) -> decimal.Decimal:
        budget: domain.Budget = self.budget_repo.get_by_id(cmd.budget_id)
        cash_flow: decimal.Decimal = budget.calculate_cash_flow_per_month()
        cash_flow = self.exchange_rate_svc.convert_currency(
            cash_flow, budget.currency, cmd.currency
        )
        return cash_flow



import logging.config

import pydoca
import tests.integration.test_app.budget.providers as providers


class Configuration(pydoca.AdaptersConfig):
    BudgetRepository = providers.InMemoryBudgetRepo
    ExchangeRateService = providers.FakeExchangeRateSvc


LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
    },
    "handlers": {
        "default": {
            "level": "DEBUG",
            "formatter": "standard",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        }
    },
}
logging.config.dictConfig(LOGGING_CONFIG)



from .budget import (
    Budget as Budget,
)
from .budget import (
    Currency as Currency,
)
from .budget import (
    Expense as Expense,
)
from .budget import (
    ExpenseAdded as ExpenseAdded,
)
from .budget import (
    FrequencyPerYear as FrequencyPerYear,
)
from .budget import (
    Income as Income,
)
from .budget import (
    IncomeAdded as IncomeAdded,
)



"""Domain-Driven Design Event."""
import datetime
from typing import Any

import pydantic

from .utils import utc_now
from .value_object import ValueObject


class Event(ValueObject):
    """Represents an event in the domain.

    Events are value objects that capture a significant occurrence or
    state change in the system. They are immutable and represent facts
    that have already happened.

    Attributes:
        timestamp: The timestamp of the event (default: utc now).
    """

    timestamp: datetime.datetime = pydantic.Field(default_factory=utc_now)

    def __init__(self, **data: Any) -> None:
        if type(self) is Event:
            raise TypeError(
                "Event cannot be instantiated directly. Please subclass it and define your attributes."
            )
        super().__init__(**data)



import pydoca
from pydoca import ID


class WheelChanged(pydoca.Event):
    position: str
    new_reference: str


class Wheel(pydoca.Entity):
    position: str
    reference: str

    def _id(self) -> ID:
        return f"{self.reference}"


class Car(pydoca.AggregateRoot):
    vin: str
    wheels: list[Wheel]

    def _id(self) -> str:
        return self.vin.lower()

    def change_wheel(self, new_wheel: Wheel) -> None:
        wheel_to_replace_idx: int = 0
        for idx, wheel in enumerate(self.wheels):
            if wheel.position == new_wheel.position:
                wheel_to_replace_idx = idx

        self.wheels.pop(wheel_to_replace_idx)
        self.wheels.append(new_wheel)
        self.add_event(
            WheelChanged(
                position=new_wheel.position,
                new_reference=new_wheel.reference,
            )
        )


def test_aggregate_root_events() -> None:
    car = Car(
        vin="vin123",
        wheels=[
            Wheel(reference="ref1", position="top_left"),
            Wheel(reference="ref1", position="top_right"),
            Wheel(reference="ref1", position="bottom_left"),
            Wheel(reference="ref1", position="bottom_right"),
        ],
    )

    car.change_wheel(Wheel(reference="ref2", position="top_left"))

    assert len(car.get_events()) == 1
    assert "ref2" in [wheel.reference for wheel in car.wheels]



import pytest
import requests

import tests.integration.test_app.budget.domain as domain


@pytest.mark.order(1)
def test_create_budget(fastapi_server, port):
    payload = {
        "budget_title": "Integration Tests",
        "budget_currency": domain.Currency.CAD.value,
    }
    response = requests.post(f"http://localhost:{port}/budget", json=payload)
    assert response.status_code == 200, response.text
    assert response.json() == {
        "title": "Integration Tests",
        "currency": domain.Currency.CAD.value,
        "incomes": [],
        "expenses": [],
        "id": "integration tests",
    }


@pytest.mark.order(2)
def test_add_operations(port):
    payload = {
        "budget_id": "integration tests",
        "operations": [
            {
                "source": "Work",
                "frequency": domain.FrequencyPerYear.MONTHLY.value,
                "amount": 10000,
                "type": "income",
            },
            {
                "source": "Mortgage",
                "frequency": domain.FrequencyPerYear.MONTHLY.value,
                "amount": 5000,
                "type": "expense",
            },
        ],
    }
    response = requests.post(
        f"http://localhost:{port}/budget/integration%20tests/add_operations",
        json=payload,
    )
    assert response.status_code == 200, response.text


@pytest.mark.order(3)
def test_calculate_cashflow(port):
    response = requests.get(
        f"http://localhost:{port}/budget/integration%20tests/calculate_cashflow?currency=United%20States%20dollar"
    )
    assert response.status_code == 200, response.text
    assert response.json() == "3750.00"  #  10000 - 5000 CAD, converted to US



"""Domain-Driven Design Entity."""
import abc
from typing import Optional, Union

import pydantic

ID = Union[bytes, float, int, str]


class Entity(pydantic.BaseModel, abc.ABC):
    """Represents the core concepts of the business being model.

    An entity is an object that is defined not by its attributes, but by its identity.

    Entities encapsulate business logic and invariants, ensuring that the data within the entity remains consistent
    and adheres to the domain's rules.
    Changes to an entity's attributes are made through defined methods, allowing for controlled updates and validation.
    They represent the nouns and primary objects that matter in the domain.

    It has a lifecycle and its identity remains constant throughout its existence.
    Entities are distinguishable from other entities based on their unique ID.

    Entities can be part of Aggregates, where they are grouped together with other entities and Value Objects
    under the control of an Aggregate Root.
    This relationship helps maintain consistency and enforces boundaries within the domain model.
    Entities have a lifecycle, they can be created, updated or deleted, but their identities remain the same.
    Entities should be persisted via the Aggregate Root only.

    Properties:
        id: The unique identifier of the entity. Result of the implementation of the abstract method `_id`.

    Methods:
        _id: Abstract method to implement that must return the entity ID.
        __eq__: Checks if two entities are equal based on their IDs.
        __hash__: Returns the hash value of the entity based on its ID.
        __str__: Returns a string representation of the entity.
    """

    @pydantic.computed_field  # type: ignore  # https://github.com/python/mypy/issues/14461
    @property
    def id(self) -> ID:
        return self._id()

    @abc.abstractmethod
    def _id(self) -> ID:
        """Returns the entity ID."""

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __str__(self) -> str:
        return f"{self.__class__.__name__} {self.id}"  # type: ignore[str-bytes-safe]


class EntityError(Exception):
    """Base classe for entity errors."""

    def __init__(
        self,
        message: str,
        entity: Optional[Entity] = None,
        class_id: Optional[tuple[type[Entity], ID]] = None,
    ):
        if entity:
            message = f"{entity} {message}"
        if class_id:
            message = f"{class_id[0]} {class_id[1]} {message}"  # type: ignore[str-bytes-safe]
        super().__init__(message)


class EntityAlreadyExistError(EntityError):
    """Entity already exists."""

    def __init__(
        self,
        entity: Optional[Entity] = None,
        class_id: Optional[tuple[type[Entity], ID]] = None,
    ):
        super().__init__("already exists", entity, class_id)


class EntityNotFoundError(EntityError):
    """Entity not found."""

    def __init__(
        self,
        entity: Optional[Entity] = None,
        class_id: Optional[tuple[type[Entity], ID]] = None,
    ):
        super().__init__("not found", entity, class_id)



from ._ports import (
    BudgetRepository as BudgetRepository,
)
from ._ports import (
    ExchangeRateService as ExchangeRateService,
)
from .add_to_budget import (
    AddToBudget as AddToBudget,
)
from .add_to_budget import (
    AddToBudgetCmd as AddToBudgetCmd,
)
from .calculate_cashflow import (
    CalculateCashFlow as CalculateCashFlow,
)
from .calculate_cashflow import (
    CalculateCashflowCmd as CalculateCashflowCmd,
)
from .create_budget import (
    CreateBudget as CreateBudget,
)
from .create_budget import (
    CreateBudgetCmd as CreateBudgetCmd,
)



import abc
import decimal

import pydoca

from .. import domain


class BudgetRepository(pydoca.Repository):
    @abc.abstractmethod
    def get_by_id(self, budget_id: str) -> domain.Budget:
        """Returns a budget if it exists or raise error.

        Raises:
            EntityNotFoundError: If raise_if_not_exist is True and the budget is not found.
        """

    @abc.abstractmethod
    def save(self, budget: domain.Budget, create: bool = False) -> domain.Budget:
        """Saves the budget.

        If create is True, raise an error if already exists in the repo.
        """


class ExchangeRateService(pydoca.Service):
    @abc.abstractmethod
    def convert_currency(
        self,
        amount: decimal.Decimal,
        current_currency: domain.Currency,
        new_currency: domain.Currency,
    ) -> decimal.Decimal:
        """Converts the amount to the new currency."""



import abc
from typing import Any, ClassVar, Optional

import pydantic

from .port_adapter import Port, inject
from .unit_of_work import UnitOfWorkBase
from .value_object import ValueObject


class UnitOfWorkNotDefined(Exception):
    """When the UnitOfWork class in not defined in the Use Case."""


class Command(ValueObject):
    """Represents an intention to perform a specific action or operation within the domain.

    Encapsulates the parameters and details needed to execute a use case.
    """


class Service(Port):
    """Service interface."""


UnitOfWork = UnitOfWorkBase


class UseCase(pydantic.BaseModel):
    __uow__: ClassVar[Optional[type[UnitOfWork]]] = None
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        uow_cls: Optional[type[UnitOfWork]] = cls.__dict__.get("UnitOfWork")
        if not uow_cls:
            return

        repositories: dict[str, Any] = {
            repo_name: (repo_type, None)
            for repo_name, repo_type in uow_cls.__annotations__.items()
        }
        cls.__uow__ = pydantic.create_model(
            "UnitOfWork",
            **repositories,
            __base__=UnitOfWorkBase,
        )

    @property
    def uow(self) -> UnitOfWork:
        if not self.__uow__:
            raise UnitOfWorkNotDefined("UnitOfWork class not defined.")
        return self.__uow__()

    @pydantic.model_validator(mode="before")
    @classmethod
    def inject_providers(cls, data: Any) -> Any:
        if not data:
            data = {}
        for attr_name, attr_info in cls.model_fields.items():
            if attr_name not in data:
                # Assumes the attribute needs to be injected.
                data[attr_name] = inject(attr_info.annotation)  # type: ignore
        return data

    @abc.abstractmethod
    def exec(self, cmd: Command) -> Any:
        """Executes the Use Case."""



import pydantic
import pytest

import pydoca


class FakeValueObject(pydoca.ValueObject):
    attribute1: int
    attribute2: str


def test_value_object_immutable() -> None:
    vo = FakeValueObject(attribute1=10, attribute2="test")
    with pytest.raises(pydantic.ValidationError):
        vo.attribute1 = 20


def test_value_object_equality() -> None:
    vo1 = FakeValueObject(attribute1=10, attribute2="test")
    vo2 = FakeValueObject(attribute1=10, attribute2="test")
    vo3 = FakeValueObject(attribute1=20, attribute2="test")

    assert vo1 == vo2  # Same attributes, so they should be equal
    assert vo1 != vo3  # Different attribute value


def test_value_object_hash() -> None:
    vo1 = FakeValueObject(attribute1=10, attribute2="test")
    vo2 = FakeValueObject(attribute1=10, attribute2="test")
    vo3 = FakeValueObject(attribute1=20, attribute2="test")

    assert hash(vo1) == hash(
        vo2
    )  # Same attributes, so their hash values should be equal
    assert hash(vo1) != hash(vo3)  # Different attribute value


def test_value_object_extra_attributes() -> None:
    with pytest.raises(pydantic.ValidationError):
        FakeValueObject(attribute1=10, attribute2="test", extra_attribute=30)  # type: ignore[call-arg]


def test_direct_use_of_value_object() -> None:
    with pytest.raises(TypeError):
        pydoca.ValueObject()



"""Python Domain-Oriented Clean Architecture library."""

__version__ = "1.0.0-alpha"

from .aggregate_root import AggregateRoot as AggregateRoot
from .bootstrap import bootstrap as bootstrap
from .entity import ID as ID
from .entity import Entity as Entity
from .entity import EntityAlreadyExistError as EntityAlreadyExistError
from .entity import EntityError as EntityError
from .entity import EntityNotFoundError as EntityNotFoundError
from .event import Event as Event
from .port_adapter import Adapter as Adapter
from .port_adapter import AdapterNotConfiguredError as AdapterNotConfiguredError
from .port_adapter import AdaptersConfig as AdaptersConfig
from .port_adapter import Port as Port
from .port_adapter import PortNotFoundError as PortNotFoundError
from .port_adapter import bind as bind
from .port_adapter import clear as clear
from .port_adapter import inject as inject
from .repository import Repository as Repository
from .repository import Session as Session
from .unit_of_work import DifferentSessionsError as DifferentSessionsError
from .unit_of_work import EventBus as EventBus
from .unit_of_work import NotARepositoryError as NotARepositoryError
from .unit_of_work import UnitOfWorkBase as UnitOfWorkBase
from .use_case import Command as Command
from .use_case import Service as Service
from .use_case import UnitOfWorkNotDefined as UnitOfWorkNotDefined
from .use_case import UseCase as UseCase
from .utils import utc_now as utc_now
from .value_object import ValueObject as ValueObject



import pydoca


class TestEntity(pydoca.Entity):
    name: str

    def _id(self) -> str:
        return self.name


def test_entity_id() -> None:
    entity = TestEntity(name="test")
    assert entity.id == "test"


def test_entity_eq() -> None:
    entity1 = TestEntity(name="test")
    entity2 = TestEntity(name="test")
    entity3 = TestEntity(name="")
    assert entity1 == entity2
    assert entity1 != entity3


def test_entity_hash() -> None:
    entity = TestEntity(name="test")
    assert hash(entity) == hash("test")


def test_entity_str() -> None:
    entity = TestEntity(name="test")
    assert str(entity) == "TestEntity test"



import pydoca

from .. import domain
from . import _ports


class CreateBudgetCmd(pydoca.Command):
    budget_title: str
    budget_currency: domain.Currency


class CreateBudget(pydoca.UseCase):
    # Test repo inside uow
    class UnitOfWork:
        budget_repo: _ports.BudgetRepository

    def exec(self, cmd: CreateBudgetCmd) -> domain.Budget:
        budget = domain.Budget(title=cmd.budget_title, currency=cmd.budget_currency)
        return self.uow.budget_repo.save(budget, create=True)



import decimal

import fastapi

import tests.integration.test_app.budget.application as application
import tests.integration.test_app.budget.domain as domain

app = fastapi.FastAPI(debug=True)


@app.post("/budget")
def create_budget(payload: application.CreateBudgetCmd) -> domain.Budget:
    return application.CreateBudget().exec(payload)


@app.get("/budget/{budget_id}/calculate_cashflow")
def calculate_cashflow(budget_id: str, currency: domain.Currency) -> decimal.Decimal:
    return application.CalculateCashFlow().exec(
        application.CalculateCashflowCmd(
            budget_id=budget_id,
            currency=currency,
        )
    )


@app.post("/budget/{budget_id}/add_operations")
def add_operations(
    budget_id: str, payload: application.AddToBudgetCmd
) -> domain.Budget:
    return application.AddToBudget().exec(payload)



import multiprocessing
import random
import time

import pytest
import uvicorn


def start_api(port: int) -> None:
    uvicorn.run(
        "tests.integration.test_app.budget.actors.api:app",
        reload=False,
        log_level="debug",
        port=port,
    )


@pytest.fixture(scope="session")
def port():
    return random.randint(8000, 9000)


@pytest.fixture(scope="session")
def fastapi_server(port):
    proc = multiprocessing.Process(target=start_api, args=(port,), daemon=True)
    proc.start()
    time.sleep(5)
    yield
    proc.kill()



import abc

import pytest

import pydoca


def test_use_case_no_dependencies():
    class UseCaseNoDependency(pydoca.UseCase):
        def exec(self, cmd: pydoca.Command) -> str:
            return "result"

    assert UseCaseNoDependency().exec(pydoca.Command()) == "result"


class Dependency(pydoca.Service):
    """Define an abstract class representing a dependency."""

    @abc.abstractmethod
    def some_method(self) -> str:
        """Returns some string."""


class DependencyImpl(Dependency):
    def some_method(self) -> str:
        return "Real implementation"


def test_use_case_with_dependencies():
    pydoca.bind(Dependency, DependencyImpl)

    class UseCaseWithDependencies(pydoca.UseCase):
        service: Dependency  # Should be injected
        name: str  # Passed in the UC init, shouldn't be injected

        def exec(self, cmd: pydoca.Command) -> str:
            return self.service.some_method()

    assert (
        UseCaseWithDependencies(name="test").exec(pydoca.Command())
        == "Real implementation"
    )


def test_use_case_missing_dependencies():
    class UseCaseWithDependencies(pydoca.UseCase):
        service: Dependency

        def exec(self, cmd: pydoca.Command) -> str:
            return self.service.some_method()

    with pytest.raises(
        pydoca.AdapterNotConfiguredError,
        match="Adapter for Dependency port not configured",
    ):
        UseCaseWithDependencies().exec(pydoca.Command())
